from osw.data.mining import RegExPatternExtended

REGEX_PATTERN_LIST = [
    RegExPatternExtended(
        description="File page full page title (new format) to label",
        pattern=r"File:(OS[WL]{1}[[a-f0-9]{32})"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Label", "Suffix(es)"],  # ["OSW-ID", "Suffix(es)"]
        example_str="File:OSW11a53cdfbdc24524bf8ac435cbf65d9d.svg",
        expected_groups=["OSW11a53cdfbdc24524bf8ac435cbf65d9d", ".svg"],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with OSW-ID and two "
        "suffixes",
        pattern=r"File:([A-Za-z]+)[\- ]{1}(OS[WL]{1}[a-f0-9]{32})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*"
        r"\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Namespace", "OSW-ID", "Label", "Suffix(es)"],
        example_str="File:LabNote-OSL9568a776dcdf4b7e86f462993df12059-"
        "spreadsheet-01.luckysheet.json",
        expected_groups=[
            "LabNote",
            "OSL9568a776dcdf4b7e86f462993df12059",
            "spreadsheet-01",
            ".luckysheet.json",
        ],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with OSW-ID and one suffix",
        pattern=r"File:([A-Za-z]+)[\- ]{1}(OS[WL]{1}[a-f0-9]{32})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Namespace", "OSW-ID", "Label", "Suffix(es)"],
        example_str="File:LabNote-OSL9ee686dbcc654f4abeaf5f38b8487885-sketch-05.svg",
        expected_groups=[
            "LabNote",
            "OSL9ee686dbcc654f4abeaf5f38b8487885",
            "sketch-05",
            ".svg",
        ],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with namespace, ISC-ID and "
        "two suffixes",
        pattern=r"File:([A-Za-z]+)[\- ]{1}([A-Za-z0-9\-]+[\- ]{1}[0-9]{4})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*"
        r"\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Namespace", "ISC-ID", "Label", "Suffix(es)"],
        example_str="File:LabNote-220304-test-0002-ni-diagram-01.drawio.svg",
        expected_groups=["LabNote", "220304-test-0002", "ni-diagram-01", ".drawio.svg"],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with namespace, ISC-ID and "
        "one suffix",
        pattern=r"File:([A-Za-z]+)[\- ]{1}([A-Za-z0-9\-]+[\- ]{1}[0-9]{4})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Namespace", "ISC-ID", "Label", "Suffix(es)"],
        example_str="File:LabNote-220305-sist-0001-ni-sketch-furnace-black.svg",
        expected_groups=[
            "LabNote",
            "220305-sist-0001",
            "ni-sketch-furnace-black",
            ".svg",
        ],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with ISC-ID and two "
        "suffixes",
        pattern=r"File:([A-Za-z0-9\-]+[\- ]{1}[0-9]{4})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*"
        r"\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["ISC-ID", "Label", "Suffix(es)"],
        example_str="File:211104-sist-0003-n stamped 20211116080457.pdf.ots",
        expected_groups=["211104-sist-0003", "n stamped 20211116080457", ".pdf.ots"],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with ISC-ID and "
        "one suffix",
        pattern=r"File:([A-Za-z0-9\-]+[\- ]{1}[0-9]{4})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["ISC-ID", "Label", "Suffix(es)"],
        example_str="File:220304-test-0003-ni stamped 20220304155047.pdf",
        expected_groups=["220304-test-0003", "ni stamped 20220304155047", ".pdf"],
    ),
    RegExPatternExtended(
        description="File page full page title with two suffixes",
        pattern=r"File:(.+)(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*"
        r"\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Label", "Suffix(es)"],
        example_str="File:OpenSemanticLab-vertical-process-abstraction-en.drawio.svg",
        expected_groups=[
            "OpenSemanticLab-vertical-process-abstraction-en",
            ".drawio.svg",
        ],
    ),
    RegExPatternExtended(
        description="File page full page title with one suffix",
        pattern=r"File:(.+)(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Label", "Suffix(es)"],
        example_str="File:ISC-Digital-1440x448px.jpg",
        expected_groups=["ISC-Digital-1440x448px", ".jpg"],
    ),
]
REGEX_PATTERN = {rep.description: rep.dict() for rep in REGEX_PATTERN_LIST}
REGEX_PATTERN_LIB = {rep.description: rep for rep in REGEX_PATTERN_LIST}



from importlib import reload
from pathlib import Path

import mwclient

import osw.model.entity as model
from osw.core import OSW
from osw.wtsite import WtSite

DEPENDENCIES = {
    "Entity": "Category:Entity",
    "Item": "Category:Item",
    "WikiFile": "Category:OSW11a53cdfbdc24524bf8ac435cbf65d9d",
}


def fetch_dependencies(wtsite_obj: WtSite):
    osw = OSW(site=wtsite_obj)
    categories_fpt = DEPENDENCIES.values()
    for i, cat in enumerate(categories_fpt):
        mode = "append"
        # if i == 0:
        #     mode = "replace"
        osw.fetch_schema(OSW.FetchSchemaParam(schema_title=cat, mode=mode))


def check_dependencies():
    dependencies_met = True
    for key in DEPENDENCIES.keys():
        if not hasattr(model, key):
            dependencies_met = False
            break
    return dependencies_met


def main(wiki_domain: str, wiki_username: str, wiki_password: str):
    dependencies_met = check_dependencies()
    if not dependencies_met:
        # For local testing without tox
        if wiki_domain is None or wiki_domain == "None":
            # Make sure that the password file is available
            cwd = Path(__file__).parent.absolute()
            pw_fp = cwd.parents[1] / "examples" / "accounts.pwd.yaml"
            wtsite = WtSite.from_domain(
                domain="wiki-dev.open-semantic-lab.org",
                password_file=str(pw_fp),
            )
        # For testing with tox
        else:
            site = mwclient.Site(host=wiki_domain)
            site.login(username=wiki_username, password=wiki_password)
            wtsite = WtSite(WtSite.WtSiteLegacyConfig(site=site))

        fetch_dependencies(wtsite_obj=wtsite)


if __name__ == "__main__":
    main()
    reload(model)



"""This script is meant to help with the migration of file pages from previous
versions to the current version of the OSW (xx and above)."""
import json
import uuid as uuid_module
from pathlib import Path
from typing import Dict, List

import mwclient
from regex_pattern import REGEX_PATTERN_LIB, REGEX_PATTERN_LIST

import osw.data.import_utility as iu
import osw.wiki_tools as wt
from osw.core import OSW
from osw.data.mining import match_first_regex_pattern, test_regex_pattern
from osw.model.entity import Label, WikiFile
from osw.utils.util import parallelize
from osw.wtsite import WtPage, WtSite

# Constants
PWD_FILE_PATH = Path(r"C:\Users\gold\ownCloud\Personal") / "accounts.pwd.yaml"
WIKIFILE_PAGE_NAME = "OSW11a53cdfbdc24524bf8ac435cbf65d9d"
WIKIFILE_NS = uuid_module.UUID(WIKIFILE_PAGE_NAME[3:])

LABEL_REGEX_PATTERNS = [
    REGEX_PATTERN_LIB[key]
    for key in [
        "File page full page title (new format) to label",
        "File page full page title (old format) with OSW-ID and two suffixes",
        "File page full page title (old format) with OSW-ID and one suffix",
        "File page full page title (old format) with namespace, ISC-ID and two "
        "suffixes",
        "File page full page title (old format) with namespace, ISC-ID and one suffix",
        "File page full page title (old format) with ISC-ID and two suffixes",
        "File page full page title (old format) with ISC-ID and one suffix",
        "File page full page title with two suffixes",
        "File page full page title with one suffix",
    ]
]
FILENAME_REGEX_PATTERNS = [
    REGEX_PATTERN_LIB[key]
    for key in [
        "File page full page title with two suffixes",
        "File page full page title with one suffix",
    ]
]


# Functions
def get_label_and_suffix_from_page_title(page_title: str) -> Dict[str, str]:
    match_res = match_first_regex_pattern(
        patterns=LABEL_REGEX_PATTERNS,
        strings=page_title,
    )
    id_ = ""
    label = ""
    suffix = ""
    if match_res[0].match is None:
        pass
    else:
        label = match_res[0].groups["Label"]
        suffix = match_res[0].groups["Suffix(es)"]
        for key in ["ISC-ID", "OSW-ID"]:
            val = match_res[0].groups.get(key, None)
            if val is not None:
                id_ = val
                break
    return {"label": label, "suffix": suffix, "id": id_}


def get_file_info_and_usage_single(
    wtsite_obj: WtSite, page_title: str, debug: bool = False
) -> dict:
    result = wt.get_file_info_and_usage(
        site=wtsite_obj._site,
        title=wt.SearchParam(
            query=page_title,
            debug=debug,
        ),
    )[0]
    other_info = get_label_and_suffix_from_page_title(page_title=page_title)
    result["info"]["label"] = other_info["label"]
    result["info"]["suffix"] = other_info["suffix"]
    result["info"]["id"] = other_info["id"]
    return result


def get_file_info_and_usage_extended(
    wtsite_obj: WtSite,
    page_titles: List[str],
    debug: bool = False,
    parallel: bool = True,
) -> List[dict]:
    result = wt.get_file_info_and_usage(
        site=wtsite_obj._site,
        title=wt.SearchParam(
            query=page_titles,
            debug=debug,
            parallel=parallel,
        ),
    )
    if len(result) != len(page_titles):
        raise ValueError(
            f"Number of results ({len(result)}) does not match number of "
            f"page titles ({len(page_titles)})"
        )
    for i, page_tile in enumerate(page_titles):
        other_info = get_label_and_suffix_from_page_title(page_title=page_tile)
        result[i]["info"]["label"] = other_info["label"]
        result[i]["info"]["suffix"] = other_info["suffix"]
        result[i]["info"]["id"] = other_info["id"]
    return result


def move_file_pages_and_redirect_references(
    file_pages: List[str],
    wtsite_obj: WtSite,
    parallel: bool = True,
    debug: bool = False,
):
    def handle_single_using_page(
        full_page_title_: str,
        wtsite_obj_: WtSite,
        file_name_: str,
        uuid_: uuid_module.UUID,
    ):
        """Adds an uuid reference to the template of the using page's (main slot)
        content."""
        file_name_with_underscores_ = file_name_.replace(" ", "_")
        using_page = WtPage(wtSite=wtsite_obj_, title=full_page_title_)
        main_slot_content = using_page.get_slot_content(slot_key="main")
        parsed_content = wt.create_flat_content_structure_from_wikitext(
            main_slot_content
        )
        replaced = False
        # Loop through lines of the using page's content
        for ii, list_item in enumerate(parsed_content):
            if isinstance(list_item, dict):
                # Loop through template dict
                for template_key, template_val in list_item.items():
                    if isinstance(template_val, dict):
                        # Look for file_name key
                        # template_fn is of type list!
                        template_fn = template_val.get("file_name", None)
                        template_img = template_val.get("image", None)
                        # In case of a match, add uuid key with uuid value (new page
                        #  uuid)
                        if template_fn is not None:
                            if (
                                file_name_ in template_fn
                                or file_name_with_underscores_ in template_fn
                            ):
                                parsed_content[ii][template_key]["uuid"] = str(uuid_)
                                replaced = True
                        if template_img is not None:
                            if (
                                file_name_ in template_img
                                or file_name_with_underscores_ in template_img
                            ):
                                parsed_content[ii][template_key]["uuid"] = str(uuid_)
                                replaced = True
        modified_content = wt.get_wikitext_from_flat_content_structure(parsed_content)
        using_page.set_slot_content(slot_key="main", content=modified_content)
        using_page.edit(comment="Edited by bot to comply with new Wiki-File schema")
        return replaced

    def handle_single_file_page(file_info_: dict, wtsite_obj_: WtSite):
        info = file_info_["info"]
        usage = file_info_["usage"]
        id_str = info["id"]
        # Don't continue if there is no usage!
        if len(usage) != 0:
            page_title = file_info_["info"]["title"]
            file_page = WtPage(wtSite=wtsite_obj_, title=page_title)
            match_res = match_first_regex_pattern(
                patterns=REGEX_PATTERN_LIST,
                strings=page_title,
            )
            if match_res[0].match is None:
                raise ValueError(
                    f"Could not match page title '{page_title}' to any of the "
                    f"supported patterns"
                )
            else:
                file_name = match_res[0].groups["Label"]

            if len(usage) == 0:
                creation_context = ""
            else:
                creation_context = usage[0]
            if id_str != "":
                # try to find a page with the same id in the usage list
                for full_page_title in usage:
                    if id_str in full_page_title:
                        creation_context = full_page_title
                        break
            # Generate UUID for the new page name
            rnd_uuid = uuid_module.uuid4()
            uuid = uuid_module.uuid5(namespace=WIKIFILE_NS, name=str(rnd_uuid))
            # Generate full page title from UUID
            new_fpt_without_suffix = iu.uuid_to_full_page_title(
                uuid=uuid, wiki_ns="File", prefix="OSW"
            )
            new_fpt = new_fpt_without_suffix + info["suffix"]

            # Loop through using pages - check if the script covers the use case of
            #  at least on using page
            covered = False
            for full_page_title in usage:
                ret_val = handle_single_using_page(
                    full_page_title_=full_page_title,
                    wtsite_obj_=wtsite_obj_,
                    file_name_=file_name,
                    uuid_=uuid,
                )
                if ret_val:
                    covered = True
            # Only move the page if the use case was covered
            if covered:
                # Rename the page
                # file_page.move(new_title=new_fpt, redirect=False)
                try:
                    file_page.move(new_title=new_fpt, redirect=False)
                except mwclient.APIError as e:
                    if e.code == "badtoken":
                        print("Got badtoken, retrying")
                        # wtsite_obj_.clear_token_cache()
                        file_page.move(new_title=new_fpt, redirect=False)  # retry

                # Make sure the slot content to be set is in accordance with the JSON
                #  schema of the category
                wiki_file_instance = WikiFile(
                    uuid=uuid,
                    label=[Label(text=info["label"])],
                    creator=[info["author"]],
                    creation_context=[creation_context],
                    # todo: new filename here? -> Name?
                    name=iu.create_page_name_from_label(label=info["label"]),
                    editor=info["editor"],
                    editing_context=usage,
                )
                # Get jsondata for the new wiki file page from the WikiFile instance
                #  with dict() or json()
                jsondata = json.loads(wiki_file_instance.json())
                file_page.set_slot_content(slot_key="jsondata", content=jsondata)
                file_page.set_slot_content(
                    "header", "{{#invoke:Entity|header}}"
                )  # required for json parsing and header rendering
                file_page.set_slot_content(
                    "footer", "{{#invoke:Entity|footer}}"
                )  # required for footer rendering
                # Actually edit the page in the OSL instance
                file_page.edit(
                    comment="Edited by bot to comply with new Wiki-File schema"
                )
                # # Alternatively, use the following to edit the page in the OSL
                # #  instance
                # osw_obj = OSW(site=wtsite_obj_)
                # osw_obj.store_entity(OSW.StoreEntityParam(
                #     entities=[wiki_file_instance], namespace="File"
                # ))
            return new_fpt

    # This is executed in parallel first (!)
    file_infos = get_file_info_and_usage_extended(
        wtsite_obj=wtsite_obj,
        page_titles=file_pages,
        debug=debug,
        parallel=parallel,
    )
    # Loop through file_pages
    if parallel:
        result = parallelize(
            handle_single_file_page,
            file_infos,
            wtsite_obj_=wtsite_obj,
            flush_at_end=debug,
        )
    else:
        result = []
        for fi in file_infos:
            result.append(
                handle_single_file_page(file_info_=fi, wtsite_obj_=wtsite_obj)
            )
    return result


def get_wiki_file_pages(wtsite_obj: WtSite, limit: int = 1000) -> List[str]:
    full_page_titles = OSW(site=wtsite_obj).query_instances(
        category=OSW.QueryInstancesParam(
            categories=[WIKIFILE_PAGE_NAME],
            limit=limit,
        )
    )
    return full_page_titles


def validation():
    patterns_to_test = LABEL_REGEX_PATTERNS
    wtsite = WtSite.from_domain(
        domain="wiki-dev.open-semantic-lab.org",
        password_file=PWD_FILE_PATH,
    )
    wiki_file_pages = get_wiki_file_pages(wtsite_obj=wtsite, limit=5000)  # replace with
    # OSW.get_instances_of(Cat:WIkiFile)
    file_pages_ = wtsite.get_file_pages()
    file_pages_to_modify = list(set(file_pages_) - set(wiki_file_pages))
    print("Collecting info and usage for file pages...")
    file_infos = get_file_info_and_usage_extended(
        wtsite_obj=wtsite,
        page_titles=file_pages_to_modify,
        debug=False,
        parallel=True,
    )

    regex_test = test_regex_pattern(patterns_to_test, file_pages_to_modify)
    return file_infos, regex_test


def main(number_of_pages_to_process: int = None):
    wtsite = WtSite.from_domain(
        domain="wiki-dev.open-semantic-lab.org",
        password_file=PWD_FILE_PATH,
    )
    wiki_file_pages = get_wiki_file_pages(wtsite_obj=wtsite, limit=5000)  # replace with
    # OSW.get_instances_of(Cat:WIkiFile)
    file_pages_ = wtsite.get_file_pages()
    file_pages_to_modify = list(set(file_pages_) - set(wiki_file_pages))
    print(f"Number of file pages to modify: {len(file_pages_to_modify)}")
    if number_of_pages_to_process is None:
        move_file_pages_and_redirect_references(
            file_pages_to_modify, wtsite, parallel=True
        )
    else:
        move_file_pages_and_redirect_references(
            file_pages_to_modify[:number_of_pages_to_process], wtsite, parallel=True
        )


if __name__ == "__main__":
    main()



from pybars import Compiler


def compile_handlebars_template(template):
    """compiles a handlebars template.
    WARNING: not thread safe!

    Parameters
    ----------
    template
        the template string

    Returns
    -------
        the compiled template
    """
    compiler = Compiler()
    compiled_template = compiler.compile(template)
    return compiled_template


def eval_compiled_handlebars_template(
    compiled_template, data, helpers=None, partials=None, add_self_as_partial=True
):
    """evaluates a compiled handlebars template with the given data

    Parameters
    ----------
    compiled_template
        the compiled template
    data
        the data dictionary
    helpers, optional
        helper functions, by default {}
    partials, optional
        partials, by default {}
    add_self_as_partial, optional
        if true, add the compiled template as partial 'self', by default True

    Returns
    -------
        the evaluated template as a string
    """
    if helpers is None:
        helpers = {}
    if partials is None:
        partials = {}
    if add_self_as_partial:
        partials["self"] = compiled_template
    return compiled_template(data, helpers=helpers, partials=partials)


def eval_handlebars_template(
    template, data, helpers={}, partials={}, add_self_as_partial=True
):
    """evaluates a handlebars template with the given data.
    WARNING: not thread safe!

    Parameters
    ----------
    template
        the template string
    data
        the data dictionary
    helpers, optional
        helper functions, by default {}
    partials, optional
        partials, by default {}
    add_self_as_partial, optional
        if true, add the compiled template as partial 'self', by default True

    Returns
    -------
        the evaluated template as a string
    """
    compiled_template = compile_handlebars_template(template)
    return eval_compiled_handlebars_template(
        compiled_template,
        data,
        helpers,
        partials,
        add_self_as_partial,
    )



import os
from pathlib import Path
from typing import Dict, List, Optional, Union

from pydantic import BaseModel


class PagePackagePageSlot(BaseModel):
    url: Optional[str]
    urlPath: Optional[str]


class PagePackagePage(BaseModel):
    """Dataclass for Page.
    Attributes are partly already defined in the specification of
    https://www.mediawiki.org/wiki/Extension:Page_Exchange"""

    name: str
    """The name (minus of the namespace) of a wiki page."""
    label: Optional[str]
    """The display label of the page."""
    namespace: Union[str, int]
    """The namespace code for the namespace of the wiki page, though stored as
    a string (like "NS_TEMPLATE"). The default is "NS_MAIN"."""
    url: Optional[str]
    """The URL at which the contents of the page can be found. This can be any URL, and
    does not have to be part of a wiki. If you are using a MediaWiki wiki page for the
    URL, make sure that the URL ends with "?action=raw" (or "&action=raw"), so that only
    the actual wikitext for the page is retrieved."""
    urlPath: Optional[str]
    """Similar to url, but gets appended to the baseURL value set for the package, if
    one was set."""
    # todo: include originURL or at PagePackage or Bundle level?
    fileURL: Optional[str]
    """ If this is a file/image page (i.e. with a "namespace" value of "NS_FILE"), this
    parameter holds the actual file, while the "url" parameter holds a URL containing
    the text contents of the wiki page."""
    fileURLPath: Optional[str]
    """Similar to fileURL, but gets appended to the baseURL value set for the package,
    if one was set."""
    slots: Dict[str, PagePackagePageSlot]
    """Slots used to store data, e.g., the main slot, storing the wiki text of the
    page. Other slots could be jsondata, jsonschema etc."""


class PagePackageNamespaceSettings(BaseModel):
    """The group of settings, per namespace, that together dictate the structure of the
    directory.
    Attributes are partly already defined in the specification of
    https://www.mediawiki.org/wiki/Extension:Page_Exchange"""

    namespace: str
    """The namespace code for this namespace, though stored as a string (like
    "NS_TEMPLATE")."""
    fileNamePrefix: Optional[str]
    """The prefix (potentially including slashes) for any file that corresponds to
    a page in this namespace."""
    fileNameSuffix: Optional[str]
    """The suffix (if any) for any file that corresponds to a page in this namespace."""
    actualFileNamePrefix: Optional[str]
    """ Used only for the "NS_FILE" namespace; specifies the prefix for the actual files
    (e.g., the images)."""
    actualFileNameSuffix: Optional[str]
    """Used only for the "NS_FILE" namespace; specifies the suffix for the actual
    files."""


class PagePackageDirectoryStructure(BaseModel):
    """For pages stored within a single repository (like GitHub), specifies the
    structure of that repository, instead of (or in addition to) any pages explicitly
    listed.
    Attributes are partly already defined in the specification of
    https://www.mediawiki.org/wiki/Extension:Page_Exchange"""

    service: Optional[str] = "GitHub"
    """The service being used to hold the data on individual pages. Currently only
    value is supported for this: "GitHub"."""
    accountName: str
    """The name of the account, i.e. the GitHub username of the repository."""
    repositoryName: str
    """The name of the specific repository."""
    namespaceSettings: PagePackageNamespaceSettings
    """The group of settings, per namespace, that together dictate the structure of the
    directory."""


class PagePackage(BaseModel):
    """A package of pages.
    Attributes are partly already defined in the specification of
    https://www.mediawiki.org/wiki/Extension:Page_Exchange"""

    globalID: str
    """(mandatory) A text identifier for this package, which is ideally globally
    unique. Ideally it uses reverse domain name notation. For example, for a package
    in the Spanish language for a CRM data structure, created by a company called
    Acme, whose internet domain is acme.com, the package identifier could be
    "com.acme.CRM.es". You can feel free to be creative within this system. For
    example, if you are publishing a package on your own and do not have your own
    internet domain, but you do have a username on mediawiki.org ("Joey User"),
    then you could give such a package the identifier
    "org.mediawiki.user.Joey_User.CRM.es". The important thing is uniqueness. It's
    also important, once people have started using a package, not to change its
    global ID - changing it would prevent people who have already downloaded the
    package from updating to more recent versions."""
    label: Optional[str]
    """The display label of the package."""
    description: str
    """A description of the page package."""
    publisher: Optional[str]
    """The default publisher."""
    publisherURL: Optional[str]
    """The default URL for the publisher."""
    author: Optional[List[str]]
    """The default author."""
    language: Optional[str]
    """The default language for all defined packages, using the (usually) two-letter
    IETF language tag for that language."""
    url: Optional[str]
    """A URL for a web page describing this entire package, if one exists."""
    version: str
    """A version number for the package, which can be updated so that users know if
    their local copy of the package is out of date."""
    licenseName: Optional[str]
    """The default license under which these packages are published."""
    requiredExtensions: Optional[List[str]]
    """An array of the names of any extensions required for this package to work. (In
    the future, this parameter may also allow defining specific versions that are
    required for specific extensions, but this is not currently possible.)"""
    requiredPackages: Optional[List[str]]
    """An array of the names of any additional packages required by this package."""
    baseURL: str
    """Holds a URL fragment (like 'https://example.com/packages/') that should be
    prepended to the urlPath or fileURLPath values set for any individual pages."""
    pages: Optional[List[PagePackagePage]]
    """The set of pages in this package."""
    directoryStructure: Optional[PagePackageDirectoryStructure]
    """For pages stored within a single repository (like GitHub), specifies the
    structure of that repository, instead of (or in addition to) any pages explicitly
    listed."""


class PagePackageBundle(BaseModel):
    """Bundle of page packages.
    Attributes are partly already defined in the specification of
    https://www.mediawiki.org/wiki/Extension:Page_Exchange"""

    publisher: Optional[str]
    """The default publisher."""
    publisherURL: Optional[str]
    """The default publisherURL."""
    author: Optional[List[str]]
    """The default author."""
    language: Optional[str] = "en"
    """The default language for all defined packages, using the
    (usually) two-letter IETF language tag for that language."""
    licenseName: Optional[str] = "CC BY-NC 4.0"
    """The default license under which these packages are published."""
    packages: Dict[str, PagePackage]
    """Holds the set of packages, with the package name as the key
    and the set of package parameters as the values."""


class PagePackageConfig(BaseModel):
    """A config for a page package"""

    name: str
    """The name (label) of the package."""
    config_path: Union[str, Path]
    """The path of the generated json file (package.json).
    Will be created automatically if not existing"""
    content_path: Optional[Union[str, Path]] = ""
    """The directory where the content (pages, files) is stored.
    Will be created automatically if not existing."""
    titles: List[str]
    """List of page titles."""
    # replace: Optional[bool] = False
    bundle: PagePackageBundle
    """Bund of pages."""
    skip_slot_suffix_for_main: Optional[bool] = False
    include_files: Optional[bool] = True

    def __init__(self, **data):
        """Originally, the dataclass.__post_init_() method was used here.
        Pydantic does not implement anything similar yet. According to
        https://github.com/pydantic/pydantic/issues/1729#issuecomment-1300576214
        no __post_init__ method will be available until Pydantic v2. Then an
        analogous function will be implemented and any in a BaseModel class defined
        method named 'model_post_init(self, **kwargs)' will be called after
        init."""
        super().__init__(**data)
        if self.content_path == "":
            self.content_path = os.path.dirname(self.config_path)


class PagePackageMetaData(BaseModel):
    """Metadata for a page package. This data needed to create a page package and
    included in the page package."""

    name: str
    """The name (label) of the page package."""
    repo: str
    """Page package repository name - usually the GitHub repository name"""
    id: str
    """Page package ID - usually the same as package_repo"""
    subdir: str
    """Page package subdirectory - usually resembling parts of the package name"""
    branch: str
    """Page package branch - usually 'main'."""
    repo_org: str
    """(GitHub) Organization hosting the package repository"""
    description: str
    """Page package description"""
    language: str = "en"
    """Page package language - usually the two-letter IETF language tag for that
    language"""
    version = "0.2.1"
    """Page package version - use semantic versioning"""
    author: List[str]
    """Author(s) of the page package"""
    publisher: str
    """Publisher of the page package."""
    page_titles: List[str]
    """List of the page titles (full page titles with namespace, e.g. 'Category:Entity')
     to be packaged."""

    requiredExtensions: Optional[List[str]]
    """The default value for PagePackage.requiredExtensions.
    An array of the names of any extensions required for this package to work. (In
    the future, this parameter may also allow defining specific versions that are
    required for specific extensions, but this is not currently possible.)"""
    requiredPackages: Optional[List[str]]
    """The default value for PagePackage.requiredPackages.
    An array of the names of any additional packages required by this package."""


# Special namespace mappings (default is namespace_const = "NS_" + namespace.upper())
NAMESPACE_CONST_TO_NAMESPACE_MAPPING = {
    # SMW: https://github.com/SemanticMediaWiki/SemanticMediaWiki/blob/ebb03c1537810f4ee8c1a25198b8d2e243cc38a1/src/NamespaceManager.php#L119
    "SMW_NS_PROPERTY": "Property",
    "SMW_NS_PROPERTY_TALK": "Property_talk",
    "SMW_NS_CONCEPT": "Concept",
    "SMW_NS_CONCEPT_TALK": "Concept_talk",
    "SMW_NS_SCHEMA": "smw/schema",
    "SMW_NS_SCHEMA_TALK": "smw/schema_talk",
    "SMW_NS_RULE": "Rule",
    "SMW_NS_RULE_TALK": "Rule_talk",
}
# inverse
NAMESPACE_TO_NAMESPACE_CONST_MAPPING = {
    v: k for k, v in NAMESPACE_CONST_TO_NAMESPACE_MAPPING.items()
}



from osw.sparql_client_smw import SmwSparqlClient

domain = "onterface.open-semantic-lab.org"


def test_sparql_client():
    sc = SmwSparqlClient(
        f"https://graph.{domain}/blazegraph/namespace/kb/sparql", domain
    )
    result = sc.get_sparql_triplets(
        where_statement=f"?subject swivt:page <https://{domain}/wiki/Main_Page>"
    )
    result_value = result["wiki:Main_Page"]["rdfs:isDefinedBy"][0]
    expected_value = f"https://{domain}/wiki/Special:ExportRDF/Main_Page"
    assert result_value == expected_value



import functools
import os
from typing import IO, List, Optional

from osw.controller.file.base import FileController
from osw.controller.file.remote import RemoteFileController
from osw.core import OSW, model
from osw.utils.wiki import get_namespace, get_title
from osw.wtsite import WtSite


class WikiFileController(model.WikiFile, RemoteFileController):
    """File controller for wiki files"""

    label: Optional[List[model.Label]] = [model.Label(text="Unnamed file")]
    """the label of the file, defaults to 'Unnamed file'"""
    osw: OSW
    """the OSW instance to connect with"""
    namespace: Optional[str] = "File"
    """the namespace of the file, defaults to 'File'"""
    title: Optional[str] = None
    """the title of the file, defaults to the auto-generated OSW-ID"""
    suffix: Optional[str] = None
    """the suffix of the file, defaults to the suffix of the title"""

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def from_other(cls, other: FileController, osw: OSW) -> "WikiFileController":
        return super().from_other(other, osw=osw)

    def get(self) -> IO:
        self._init()
        file = self.osw.site._site.images[self.title]
        # return file.download() # in-memory - limited by available RAM

        response = self.osw.site._site.connection.get(
            file.imageinfo["url"], stream=True
        )
        # for chunk in response.iter_content(1024):
        #    destination.write(chunk)
        # see https://stackoverflow.com/questions/16694907/download-large-file-in-python-with-requests
        response.raw.read = functools.partial(response.raw.read, decode_content=True)
        return response.raw

    def get_to(self, other: "FileController"):
        self._init()
        file = self.osw.site._site.images[self.title]
        # return file.download() # in-memory - limited by available RAM

        with self.osw.site._site.connection.get(
            file.imageinfo["url"], stream=True
        ) as response:
            # for chunk in response.iter_content(1024):
            #    destination.write(chunk)
            # see https://stackoverflow.com/questions/16694907/download-large-file-in-python-with-requests
            response.raw.read = functools.partial(
                response.raw.read, decode_content=True
            )
            other.put(response.raw)

    def put(self, file: IO):
        # extract file meta information
        if hasattr(file, "name") and file.name is not None:
            name = os.path.basename(file.name)
            suffix = ""
            if "." in name:
                suffix = "." + name.split(".")[-1]
            self._init(name, suffix)

        # file_page = self.osw._site.get_page(WtSite.GetPageParam(titles=[file_page_name])).pages[0]
        self.osw.store_entity(
            OSW.StoreEntityParam(
                entities=[self.cast(model.WikiFile)], namespace=self.namespace
            )
        )
        self.osw.site._site.upload(
            file=file,
            filename=self.title,
            # comment="",
            # description="",
            ignore=True,
        )

    def put_from(self, other: FileController):
        # if isinstance(file, LocalFileController) and self.suffix is None:
        #    lf = file.cast(LocalFileController)
        #    self.meta.wiki_page.title.removesuffix(lf.path.suffix)
        #    self.meta.wiki_page.title += lf.path.suffix
        # copy over metadata
        if self.label == [model.Label(text="Unnamed file")]:
            self.label = other.label
        super().put_from(other)

    def delete(self):
        file_page_name = f"{self.namespace}:{self.title}"
        file_page = self.osw.site.get_page(
            WtSite.GetPageParam(titles=[file_page_name])
        ).pages[0]
        if file_page.exists:
            file_page.delete()

    def _init(self, name=None, suffix=None):
        # set the name attribute to the actual file name, e. g. "image.png"
        if self.name is None:
            self.name = name
        if suffix is None:
            title = get_title(self)
            if "." in title:
                suffix = "." + title.split(".")[-1]
            else:
                suffix = ""
        if self.suffix is None:
            self.suffix = suffix
        self.name = self.name.removesuffix(self.suffix)
        # set the title from OSW-ID + suffix, e. g. "OSWeedfa07ad404421b8622e0099624d254.png"
        if self.title is None:
            self.title = get_title(self)
            self.title = self.title.removesuffix(self.suffix)
            self.title += self.suffix
        if self.namespace is None:
            self.namespace = get_namespace(self)
        # update meta information
        if not hasattr(self, "meta"):
            self.meta = model.Meta()
        elif self.meta is None:
            self.meta = model.Meta()
        if not hasattr(self.meta, "wiki_page") or self.meta.wiki_page is None:
            self.meta.wiki_page = model.WikiPage()
        self.meta.wiki_page.title = self.title
        self.meta.wiki_page.namespace = self.namespace



import os

from osw.auth import CredentialManager
from osw.core import OSW, model
from osw.ontology import ImportConfig, OntologyImporter
from osw.wtsite import WtSite

# use credentials from file. if none are found, the user will be prompted to enter them
cm = CredentialManager(
    cred_filepath=os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
    )
)

# create the site object
wtsite = WtSite(
    WtSite.WtSiteConfig(iri="https://wiki-dev.open-semantic-lab.org", cred_mngr=cm)
)
osw = OSW(site=wtsite)

list_of_schemas = [
    "Category:OSW725a3cf5458f4daea86615fcbd0029f8",  # OwlClass
    "Category:OSW6b9ef2784a934b8ab96523366e23e906",  # OwlIndividual
    "Category:Item",
    "Category:ObjectProperty",
    "Category:DataProperty",
    "Category:AnnotationProperty",
]
for i, cat in enumerate(list_of_schemas):
    mode = "append"
    if i == 0:
        mode = "replace"
    osw.fetch_schema(OSW.FetchSchemaParam(schema_title=cat, mode=mode))

ontology_name = "example_ontology"

ex = model.Ontology(
    name="Example",
    iri="http://example.com",
    prefix="http://example.com/",
    prefix_name="example",
    link="http://example.com",
)

import_config = ImportConfig(
    ontology_name=ontology_name,
    ontologies=[ex],
    file=os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "data", f"{ontology_name}.ttl"
    ),
    base_class=model.OwlClass,
    base_class_title="Category:OSW725a3cf5458f4daea86615fcbd0029f8",  # OwlClass
    dump_files=True,
    dump_path=os.path.dirname(os.path.abspath(__file__)),
    dry_run=False,
)

# import ontologies
importer = OntologyImporter(
    osw=osw,
)
importer.import_ontology(import_config)



import os
from io import StringIO

from osw.auth import CredentialManager
from osw.core import OSW
from osw.wtsite import WtSite

# credential manager
# can use a file or hardcode the credentials, otherwise the user will be prompted to enter them
cm = CredentialManager(
    cred_filepath=os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
    )
    # CredentialManager.UserPwdCredential(
    #     iri=wiki_domain, username=wiki_username, password=wiki_password
    # )
)
# change the domain to your osw instance
osw_obj = OSW(
    site=WtSite(WtSite.WtSiteConfig(iri="wiki-dev.open-semantic-lab.org", cred_mngr=cm))
)

# load the required schemas / data classes
osw_obj.fetch_schema(
    OSW.FetchSchemaParam(
        schema_title=[
            "Category:OSW11a53cdfbdc24524bf8ac435cbf65d9d",  # WikiFile
            "Category:OSW3e3f5dd4f71842fbb8f270e511af8031",  # LocalFile
        ],
        mode="replace",
    )
)

# import the controller modules
# note: since they depend on the data classes, they must be imported after the schemas are loaded
from osw.controller.file.local import (  # noqa (ignore flake8 warning)
    LocalFileController,
)
from osw.controller.file.wiki import WikiFileController  # noqa (ignore flake8 warning)

# create a local file
# with open("dummy.txt", "w") as f:
#    f.write("Hello World!")
lf = LocalFileController(path="dummy.txt")  # here an uuid already exists
lf.put(StringIO("Hello World!"))

# create a remote file (here: a wiki file)
wf = WikiFileController(osw=osw_obj)
# or cast to wiki wile to keep all common attributes
wf = lf.cast(WikiFileController, osw=osw_obj)
# which is equivalent to
wf = WikiFileController.from_other(lf, osw=osw_obj)

# upload the local file to the remote file
wf.put_from(lf)
# write new content on the fly
wf.put(StringIO("Some new content"))

# get an existing file
file = osw_obj.load_entity(f"{wf.namespace}:{wf.title}")  # the file
wf2 = file.cast(WikiFileController, osw=osw_obj)  # the file controller
lf2 = LocalFileController.from_other(wf2, path="dummy2.txt")
lf2.put_from(wf2)

# delete the files
lf.delete()
wf.delete()  # note: wf2 actually points to the same file as wf
lf2.delete()



from pathlib import Path
from typing import Union

from pydantic import FilePath

import osw.model.page_package as model
from osw.auth import CredentialManager
from osw.model import page_package as package
from osw.model.static import OswBaseModel
from osw.wtsite import WtSite


class PagePackageController(model.PagePackageMetaData):
    """Adds controller logic to a package definitions to create / update it"""

    class CreationConfig(OswBaseModel):
        """Configuration for creating a page package. This is the data needed to create
        a page package but is not included in the page package."""

        domain: str
        """A string formatted as domain"""
        credentials_file_path: Union[str, FilePath]
        """Path to a existing credentials yaml files"""
        working_dir: Union[str, Path]
        """Working directory. Will be created automatically if not existing."""
        skip_slot_suffix_for_main: bool = False

    def create(
        self,
        creation_config: CreationConfig,
    ):
        """creates or updates the package

        Parameters
        ----------
        creation_config
            see PagePackageController.CreationConfig
        """
        # Create a WtSite instance to load pages from the specified domain
        wtsite = WtSite(
            WtSite.WtSiteConfig(
                iri=creation_config.domain,
                cred_mngr=CredentialManager(
                    cred_filepath=creation_config.credentials_file_path
                ),
            )
        )
        # Create a PagePackageBundle instance
        bundle = package.PagePackageBundle(
            publisher=self.publisher,
            author=self.author,
            language=self.language,
            publisherURL=f"https://github.com/{self.repo_org}/{self.repo}",
            packages={
                f"{self.name}": package.PagePackage(
                    globalID=f"{self.id}",
                    label=self.name,
                    version=self.version,
                    requiredExtensions=self.requiredExtensions,
                    requiredPackages=self.requiredPackages,
                    description=self.description,
                    baseURL=f"https://raw.githubusercontent.com/"
                    f"{self.repo_org}/"
                    f"{self.repo}/"
                    f"{self.branch}/"
                    f"{self.subdir}/",
                )
            },
        )
        # Create a PagePackageConfig instance
        config = package.PagePackageConfig(
            name=self.name,
            config_path=Path(creation_config.working_dir) / "packages.json",
            content_path=Path(creation_config.working_dir) / self.subdir,
            bundle=bundle,
            titles=self.page_titles,
            skip_slot_suffix_for_main=creation_config.skip_slot_suffix_for_main,
        )
        # Create the page package in the working directory
        wtsite.create_page_package(WtSite.CreatePagePackageParam(config=config))



from abc import abstractmethod
from typing import IO, Any

from osw.core import model


class FileController(model.File):
    """Base class for file controllers"""

    @abstractmethod
    def get(self) -> IO:
        """Reads or downloads a file from an IO object
        Make sure to close the IO object after reading

        Returns
        -------
            the IO object of the opened file
        """
        pass

    @abstractmethod
    def put(self, file: IO):
        """Writes or uploads a file from an IO object

        Parameters
        -------
            file
                the IO object to read from
        """
        pass

    def put_from(self, other: "FileController"):
        """Writes or uploads a file from another file controller
        Automatically closes the other file controllers IO object after reading

        Parameters
        ----------
        other
            another file controller to read from
        """
        with other.get() as f:
            self.put(f)

    def get_to(self, other: "FileController"):
        """Reads or downloads a file to another file controller
        Automatically closes the this file controllers IO object after reading

        Parameters
        ----------
        other
            another file controller to write to
        """
        with self.get() as f:
            other.put(f)

    @classmethod
    def from_other(cls, other: "FileController", **kwargs: Any) -> "FileController":
        """Create an instance based on another file controller
        copying all common attributes

        Parameters
        ----------
        other
            another file controller to copy from
        kwargs
            additional attributes to be set
        """
        return other.cast(cls, **kwargs)



"""Generic extension of mwclient.site, mainly to provide multi-slot page handling and
caching OpenSemanticLab specific features are located in osw.core.OSW
"""

import json
import os
import shutil
from datetime import datetime
from pathlib import Path
from pprint import pprint
from time import sleep
from typing import Dict, List, Optional, Union

import mwclient
from jsonpath_ng.ext import parse
from pydantic import FilePath
from typing_extensions import deprecated

import osw.model.entity as model
import osw.model.page_package as package
import osw.utils.util as ut
import osw.wiki_tools as wt
from osw.auth import CredentialManager
from osw.model.static import OswBaseModel
from osw.utils.util import parallelize

# Constants
SLOTS = {
    "main": {"content_model": "wikitext", "content_template": ""},
    "header": {"content_model": "wikitext", "content_template": ""},
    "footer": {"content_model": "wikitext", "content_template": ""},
    "jsondata": {"content_model": "json", "content_template": ""},
    "jsonschema": {"content_model": "json", "content_template": ""},
    "template": {"content_model": "wikitext", "content_template": ""},
    "header_template": {"content_model": "wikitext", "content_template": ""},
    "footer_template": {"content_model": "wikitext", "content_template": ""},
    "data_template": {"content_model": "wikitext", "content_template": ""},
    "schema_template": {"content_model": "wikitext", "content_template": ""},
}


# Classes
class WtSite:
    """A wrapper class of mwclient.Site, mainly to provide multi-slot page handling and
    caching"""

    class WtSiteConfig(OswBaseModel):
        """The configuration for a WtSite instance"""

        iri: str
        """The IRI of the wiki site, typically the domain"""
        cred_mngr: CredentialManager
        """The CredentialManager to use for authentication"""
        login: Optional[str]
        """The preferred login name when multiple logins are possible (not supported
        yet)"""

    @deprecated("Use WtSiteConfig instead")
    class WtSiteLegacyConfig(OswBaseModel):
        """The legacy configuration for a WtSite instance"""

        site: mwclient.Site
        """the mwclient.Site to use for communication with the wiki"""

        class Config:
            arbitrary_types_allowed = True

    def __init__(self, config=Union[WtSiteConfig, WtSiteLegacyConfig]):
        """creates a new WtSite instance from a WtSiteConfig

        Parameters
        ----------
        config
            the WtSiteConfig or WtSiteLegacyConfig to create the WtSite from

        """

        scheme = "https"

        if isinstance(config, WtSite.WtSiteLegacyConfig):
            self._site = config.site
        else:
            cred = config.cred_mngr.get_credential(
                CredentialManager.CredentialConfig(
                    iri=config.iri, fallback=CredentialManager.CredentialFallback.ask
                )
            )
            if "//" in config.iri:
                scheme = config.iri.split("://")[0]
                config.iri = config.iri.split("://")[1]
            if isinstance(cred, CredentialManager.UserPwdCredential):
                self._site = mwclient.Site(config.iri, path="/w/", scheme=scheme)
                self._site.login(username=cred.username, password=cred.password)
            elif isinstance(cred, CredentialManager.OAuth1Credential):
                self._site = mwclient.Site(
                    config.iri,
                    path="/w/",
                    scheme=scheme,
                    consumer_token=cred.consumer_token,
                    consumer_secret=cred.consumer_secret,
                    access_token=cred.access_token,
                    access_secret=cred.access_secret,
                )
            else:
                raise ValueError("Unsupported credential type: " + str(type(cred)))
            del cred

        # The page cache is used to store pages that already have been loaded from
        #  the wiki
        self._page_cache = {}
        self._cache_enabled = False

    @classmethod
    @deprecated("Use contructor instead")
    def from_domain(
        cls,
        domain: str = None,
        password_file: Union[str, FilePath] = None,
        credentials: dict = None,
    ):
        """creates a new WtSite instance from a domain and a password file

        Parameters
        ----------
        domain, optional
            the wiki domain, by default None
        password_file, optional
            password file with credentials for the wiki site, by default None
        credentials, optional
            credentials of the wiki site, by default None

        Returns
        -------
            a new WtSite instance
        """
        if credentials is None:
            site = wt.create_site_object(domain, password_file)
        else:
            site = wt.create_site_object(domain, "", credentials)
        return cls(WtSite.WtSiteLegacyConfig(site=site))

    @classmethod
    @deprecated("Use contructor instead")
    def from_credentials(
        cls,
        credentials: Union[Dict[str, Dict[str, str]], str, FilePath],
        key: Union[str, int] = 0,
    ):
        """

        Parameters
        ----------
        credentials:
            A dictionary of credentials or a path to a credentials file, expected to
            contain keys 'username' and 'password'.
        key:
            The key of the credentials-dictionary (domain) or the index of the key.
            By default, the first key is used.

        Returns
        -------

        """

        if isinstance(credentials, str) or isinstance(credentials, Path):
            domains, accounts = wt.read_domains_from_credentials_file(credentials)
        else:
            accounts = credentials
        if isinstance(key, int):
            _domain = list(accounts.keys())[key]
        else:
            _domain = key
        _domain = _domain
        _credentials: Dict = accounts[_domain]
        # todo: research why nested typing doesn't work as expected fo the dictionary
        site = wt.create_site_object(_domain, "", _credentials)
        return cls(WtSite.WtSiteLegacyConfig(site=site))

    class GetPageParam(model.OswBaseModel):
        titles: Union[str, List[str]]
        """title string or list of title strings of the pages to download"""
        parallel: Optional[bool] = None
        """whether to download the pages in parallel or sequentially
        Defaults to True if more than 5 pages are requested, False otherwise"""
        retries: Optional[int] = 5
        """How often to retry downloading a page if an error occurs"""
        retry_delay_s: Optional[int] = 5
        """Retry delay in seconds"""
        debug: Optional[bool] = False
        """Whether to print debug messages"""

    class GetPageResult(model.OswBaseModel):
        pages: List["WtPage"]
        """List of pages that have been downloaded"""
        errors: List[Exception]
        """List of errors that occurred while downloading pages"""

        class Config:
            arbitrary_types_allowed = True  # allows to use WtPage in type hints

    def get_page(self, param: GetPageParam) -> GetPageResult:
        """Downloads a page or a list of pages from the site.

        Parameters
        ----------
        param:
            GetPageParam object
        """
        # ensure that titles is a list
        if not isinstance(param.titles, list):
            param.titles = [param.titles]
        max_index = len(param.titles)
        if param.parallel is None and max_index >= 5:
            param.parallel = True

        exeptions = []
        pages = []

        def get_page_(title, index):
            retry = 0
            wtpage = None
            while retry < param.retries:
                msg = ""
                try:
                    if self._cache_enabled and title in self._page_cache:
                        wtpage = self._page_cache[title]
                        msg += f"({index + 1}/{max_index}): Page loaded from cache."
                    else:
                        wtpage = WtPage(self, title)
                        msg += f"({index + 1}/{max_index}): Page loaded"
                        if self._cache_enabled:
                            self._page_cache[title] = wtpage
                    pages.append(wtpage)
                    break
                except Exception as e:
                    exeptions.append(e)
                    msg += str(e)
                    if retry < param.retries:
                        retry += 1
                        msg = f"({index + 1}/{max_index}): Page load failed. Retry ({retry}/{param.retries})"
                        sleep(5)
                print(msg)
            self._clear_cookies()
            return wtpage

        if param.parallel:
            _ = parallelize(get_page_, param.titles, flush_at_end=param.debug)
        else:
            _ = [get_page_(p, i) for i, p in enumerate(param.titles)]

        return self.GetPageResult(pages=pages, errors=exeptions)

    @deprecated("Use get_page instead")
    def get_WtPage(self, title: str = None):
        """Creates a new WtPage object for the given title
           and loads the page from the site if the page already exists.
           Deprecated in favor of WtSite.get_page.

        Parameters
        ----------
        title:
            title of the page to load / create

        """
        result = self.get_page(WtSite.GetPageParam(titles=title))
        return result.pages[0]

    def enable_cache(self):
        self._cache_enabled = True

    def disable_cache(self):
        self._cache_enabled = False

    def get_cache_enabled(self):
        return self._cache_enabled

    def clear_cache(self):
        del self._page_cache
        self._page_cache = {}

    def _clear_cookies(self):
        # see https://github.com/mwclient/mwclient/issues/221
        for cookie in self._site.connection.cookies:
            if "PostEditRevision" in cookie.name:
                self._site.connection.cookies.clear(
                    cookie.domain, cookie.path, cookie.name
                )

    def prefix_search(self, text: Union[str, wt.SearchParam]):
        return wt.prefix_search(self._site, text)

    def semantic_search(self, query: Union[str, wt.SearchParam]):
        return wt.semantic_search(self._site, query)

    class ModifySearchResultsParam(model.OswBaseModel):
        mode: str
        query: wt.SearchParam
        comment: str = None
        log: bool = False
        dryrun: bool = False

    def modify_search_results(
        self,
        mode: str,
        query: str,
        modify_page,
        limit=None,
        comment=None,
        log=False,
        dryrun=False,
    ):
        titles = []
        if mode == "prefix":
            titles = wt.prefix_search(self._site, query)
        elif mode == "semantic":
            titles = wt.semantic_search(self._site, query)
        if limit:
            titles = titles[0:limit]
        if log:
            print(f"Found: {titles}")
        for title in titles:
            wtpage = self.get_page(WtSite.GetPageParam(titles=[title])).pages[0]
            modify_page(wtpage)
            if log:
                print(f"\n======= {title} =======")
                print(wtpage._content)
                for slot in wtpage._slots:
                    content = wtpage.get_slot_content(slot)
                    # if isinstance(content, dict): content = json.dumps(content)
                    print(f"   ==== {title}:{slot} ====   ")
                    pprint(content)
                    print("\n")
            if not dryrun:
                wtpage.edit(comment)

    class UploadPageParam(model.OswBaseModel):
        pages: Union["WtPage", List["WtPage"]]
        parallel: Optional[bool] = False
        debug: Optional[bool] = False

        class Config:
            arbitrary_types_allowed = True

    def upload_page(
        self,
        param: Union[UploadPageParam, "WtPage", List["WtPage"]],
    ) -> None:
        """Uploads a page or a list of pages to the site.

        Parameters
        ----------
        param:
            UploadPageParam object or a WtPage object or a list of WtPage objects.
        """
        if isinstance(param, WtPage):
            param = WtSite.UploadPageParam(pages=[param])
        elif isinstance(param, list):
            param = WtSite.UploadPageParam(pages=param)

        max_index = len(param.pages)
        if max_index >= 5:
            param.parallel = True

        def upload_page_(page, index: int = None):
            # Before uploading: Check if the page is uploaded to the WtSite that is
            #  defining this method
            if page.wtSite != self:
                raise AssertionError(
                    f"The WtSite in page '{page.title}' and the "
                    f"WtSite from which this method is called from "
                    f"are not matching!"
                )
            page.edit()

            if index is None:
                print(f"Uploaded page to {page.get_url()}.")
            else:
                print(
                    f"({index + 1}/{max_index}): Uploaded page to " f"{page.get_url()}."
                )

        if param.parallel:
            _ = parallelize(upload_page_, param.pages, flush_at_end=param.debug)
        else:
            _ = [upload_page_(p, i) for i, p in enumerate(param.pages)]

    class CreatePagePackageParam(model.OswBaseModel):
        """Parameter object for create_page_package method."""

        config: package.PagePackageConfig
        """Configuration object for the page package."""
        dump_config: Optional["WtPage.PageDumpConfig"] = None
        """Configuration object for the page dump"""
        debug: Optional[bool] = True
        """If True, debug messages will be printed."""

        class Config:
            arbitrary_types_allowed = True

    def create_page_package(self, param: CreatePagePackageParam):
        """Create a page package, which is a locally stored collection of wiki pages
        and their slots, based on a configuration object.

        Parameters:
        -----------
        param:  CreatePagePackageParam
        """

        debug = param.debug
        dump_config = param.dump_config
        config = param.config

        # Clear the content directory
        try:
            if debug:
                print(f"Delete dir '{config.content_path}'")
            if os.path.exists(config.content_path):
                shutil.rmtree(config.content_path)
        except OSError as e:
            if debug:
                print("Error: %s - %s." % (e.filename, e.strerror))
        # Create a dump config
        if dump_config is None:
            dump_config = WtPage.PageDumpConfig(
                target_dir=config.content_path,
                skip_slot_suffix_for_main=config.skip_slot_suffix_for_main,
            )
        else:
            dump_config.target_dir = config.content_path
            dump_config.skip_slot_suffix_for_main = config.skip_slot_suffix_for_main

        bundle = config.bundle  # type: package.PagePackageBundle
        added_titles = []  # keep track of added pages, prevent duplicates

        if config.name not in bundle.packages:
            print(f"Error: package {config.name} does not exist in bundle")
            return
        if not bundle.packages[config.name].pages:
            bundle.packages[config.name].pages = []
        for title in config.titles:
            if title in added_titles:
                continue  # prevent duplicates
            else:
                added_titles.append(title)
            page = self.get_page(WtSite.GetPageParam(titles=[title])).pages[0]

            bundle.packages[config.name].pages.append(page.dump(dump_config))
            if config.include_files:
                for file in page._page.images():
                    if file.name in added_titles:
                        continue  # prevent duplicates
                    else:
                        added_titles.append(file.name)
                    file_page = self.get_page(
                        WtSite.GetPageParam(titles=[file.name])
                    ).pages[0]
                    bundle.packages[config.name].pages.append(
                        file_page.dump(dump_config)
                    )

        content = bundle.json(exclude_none=True, indent=4, ensure_ascii=False)
        # This will create the JSON (e.g., package.json) with the PagePackageConfig,
        #  which contains the PagePackageBundle
        file_name = f"{config.config_path}"
        with open(file_name, "w", encoding="utf-8") as f:
            f.write(content)

    class ReadPagePackageParam(model.OswBaseModel):
        """Parameter type of read_page_package."""

        storage_path: Union[str, Path]
        """The path to the directory where the page package is stored."""
        packages_info_file_name: Optional[Union[str, Path]] = None
        """The name of the file that contains the page package information. If not
            specified, the default value 'packages.json' is used."""
        selected_slots: Optional[List[str]] = None
        """A list of slots that should be read. If None, all slots are read."""
        debug: Optional[bool] = False
        """If True, debug information is printed to the console."""

    class ReadPagePackageResult(model.OswBaseModel):
        """Return type of read_page_package."""

        pages: List["WtPage"]
        """A list of WtPage objects."""

        class Config:
            arbitrary_types_allowed = True

    def read_page_package(self, param: ReadPagePackageParam) -> ReadPagePackageResult:
        """Read a page package, which is a locally stored collection of wiki pages and
        their slots' content.

        Parameters
        ----------
        param: ReadPagePackageParam

        Returns
        -------
        result: ReadPagePackageResult
        """
        # map params
        storage_path = param.storage_path
        packages_info_file_name = param.packages_info_file_name
        selected_slots = param.selected_slots
        debug = param.debug
        # Test arguments / set default value
        if not os.path.exists(storage_path):
            raise FileNotFoundError(f"Storage path '{storage_path}' does not exist.")
        if not isinstance(storage_path, Path):
            storage_path = Path(storage_path)
        if packages_info_file_name is None:
            packages_info_file_name = "packages.json"
        # Get top level content of storage path
        top_level_content = ut.list_files_and_directories(
            search_path=storage_path, recursive=False
        )
        # Try to find packages info file in storage path
        pi_fp = os.path.join(storage_path, packages_info_file_name)
        # Check if packages info file exists in storage path, otherwise use fallback
        #  option or raise error
        if os.path.exists(pi_fp) and os.path.isfile(pi_fp):
            if debug:
                print(f"Found packages info file at '{pi_fp}'.")
        else:
            if debug:
                print(
                    f"Did not find packages info file at '{pi_fp}'. Trying default "
                    f"'packages.json'."
                )
            top_level_json_files = [
                f
                for f in top_level_content["files"]
                if f.is_file() and str(f).endswith(".json")
            ]
            json_in_top_level = ut.file_in_paths(
                paths=top_level_json_files, file_name="packages.json"
            )
            if json_in_top_level["found"]:
                pi_fp = json_in_top_level["file path"]
            elif len(top_level_json_files) > 0:
                if debug:
                    print(f"Found JSON files: {top_level_json_files}. Using first one.")
                pi_fp = top_level_json_files[0]
            else:
                raise FileNotFoundError(
                    f"Error: No JSON files found in '{storage_path}'."
                )
        # Read packages info file
        with open(pi_fp, "r", encoding="utf-8") as f:
            packages_json = json.load(f)
        # Assume that the pages files are located in the subdir
        storage_path_content = ut.list_files_and_directories(
            search_path=storage_path, recursive=True
        )
        sub_dirs = top_level_content["directories"]
        if len(top_level_content["directories"]) == 0:
            # No subdirectories found, assume that the pages files are located in the
            #  top level
            sub_dirs = [storage_path]

        def get_slot_content(
            parent_dir: List[Union[str, Path]],
            url_path: str,
            files_in_storage_path: List[Path],
        ) -> Union[str, Dict]:
            for pdir in parent_dir:
                slot_path = storage_path / pdir / url_path
                if slot_path in files_in_storage_path:
                    with open(slot_path, "r", encoding="utf-8") as f:
                        file_content = f.read()
                    # Makes sure not to open an empty file with json
                    if len(file_content) > 0:
                        if url_path.endswith(".json"):
                            with open(slot_path, "r", encoding="utf-8") as f:
                                slot_data = json.load(f)
                            return slot_data
                        elif url_path.endswith(".wikitext"):
                            slot_data = file_content
                            return slot_data

        # Create WtPage objects
        pages = []
        for package_name, package_dict in packages_json["packages"].items():
            for page in package_dict["pages"]:
                namespace = page["namespace"].split("_")[-1].capitalize()
                name = page["name"]
                # Create the WtPage object
                page_obj = WtPage(wtSite=self, title=f"{namespace}:{name}")
                if "main" in selected_slots:
                    # Main slot is special
                    slot_content = get_slot_content(
                        parent_dir=sub_dirs,
                        url_path=page["urlPath"],
                        files_in_storage_path=storage_path_content["files"],
                    )
                    if slot_content is not None:
                        page_obj.set_slot_content(
                            slot_key="main",
                            content=slot_content,
                        )
                if selected_slots is None:
                    _selected_slots = page["slots"]
                else:
                    _selected_slots = {
                        slot_name: slot_dict
                        for slot_name, slot_dict in page["slots"].items()
                        if slot_name in selected_slots
                    }
                for slot_name, slot_dict in _selected_slots.items():
                    slot_content = get_slot_content(
                        parent_dir=sub_dirs,
                        url_path=slot_dict["urlPath"],
                        files_in_storage_path=storage_path_content["files"],
                    )
                    if slot_content is not None:
                        page_obj.set_slot_content(
                            slot_key=slot_name,
                            content=slot_content,
                        )
                pages.append(page_obj)
        return WtSite.ReadPagePackageResult(pages=pages)

    class UploadPagePackageParam(model.OswBaseModel):
        """Parameter class for upload_page_package method."""

        storage_path: Optional[Union[str, Path]] = None
        """The path to the storage directory.
        If 'storage_path' is not given, 'pages' must be given."""
        pages: Optional[List["WtPage"]] = None
        """A list of WtPage objects.
        If 'pages' is not given, 'storage_path' must be given."""
        debug: Optional[bool] = False
        """If True, prints debug information."""

        class Config:
            arbitrary_types_allowed = True

    def upload_page_package(self, param: UploadPagePackageParam):
        """Uploads a page package to the wiki defined by a list of WtPage objects or
        a storage path.

        Parameters
        ----------
        param : UploadPagePackageParam

        """
        storage_path = param.storage_path
        pages = param.pages
        debug = param.debug

        if storage_path and pages is None:
            raise ValueError(
                "Error: If 'storage_path' is not given, 'pages' must be given."
            )
        if pages is None:
            pages = self.read_page_package(
                WtSite.ReadPagePackageParam(storage_path=storage_path, debug=debug)
            ).pages
        for page in pages:
            page.edit()

    def get_file_pages(self, limit: int = 1000000) -> List[str]:
        """Get all file pages in the wiki"""
        full_page_titles = wt.prefix_search(
            site=self._site,
            text=wt.SearchParam(query="File:", debug=False, limit=limit),
        )
        return full_page_titles

    def get_file_info_and_usage(
        self,
        page_titles: Union[str, List[str], wt.SearchParam],
    ) -> list:
        """Get the file info and usage for one or more file pages.

        Parameters
        ----------
        page_titles:
            One or more full page titles of file pages or wiki_tools.SearchParam.

        Returns
        -------
        result:
            Dictionary with page titles as keys and nested dictionary with keys 'info'
            and 'usage'.
        """
        if isinstance(page_titles, str):
            title = wt.SearchParam(query=[page_titles], debug=False, parallel=True)
        elif isinstance(page_titles, list):
            title = wt.SearchParam(query=page_titles, debug=False, parallel=True)
        else:  # SearchParam
            title = page_titles
        return wt.get_file_info_and_usage(site=self._site, title=title)


class WtPage:
    def __init__(self, wtSite: WtSite = None, title: str = None):
        self.wtSite = wtSite
        self.title = title

        self._page = wtSite._site.pages[self.title]
        self.exists = self._page.exists
        self._original_content = ""
        self._content = ""
        self.changed: bool = False
        self._dict = []  # todo: named dict but is of type list
        self._slots: Dict[str, Union[str, dict]] = {"main": ""}
        self._slots_changed: Dict[str, bool] = {"main": False}
        self._content_model: Dict[str, str] = {"main": "wikitext"}

        if self.exists:
            self._original_content = self._page.text()
            self._content = self._original_content
            self._dict = wt.create_flat_content_structure_from_wikitext(
                self._content, array_mode="only_multiple"
            )
            # multi content revisions
            rev = wtSite._site.api(
                "query",
                prop="revisions",
                titles=title,
                rvprop="ids|timestamp|flags|comment|user|content|contentmodel|roles|"
                "slotsize|slotsha1",
                rvslots="*",
                rvlimit="1",
                format="json",
            )
            for page_id in rev["query"]["pages"]:
                page = rev["query"]["pages"][page_id]
                if page["title"] == title:
                    for revision in page["revisions"]:
                        self._current_revision = revision
                        for slot_key in revision["slots"]:
                            self._slots[slot_key] = revision["slots"][slot_key]["*"]
                            self._content_model[slot_key] = revision["slots"][slot_key][
                                "contentmodel"
                            ]
                            self._slots_changed[slot_key] = False
                            # self._slots_sha1[slot_key] = \
                            #     revision["slots"][slot_key]["*"]
                            if self._content_model[slot_key] == "json":
                                self._slots[slot_key] = json.loads(
                                    self._slots[slot_key],
                                )
                    # todo: set content for slots not in revision["slots"] (use
                    #  SLOTS) --> create empty slots

    def create_slot(self, slot_key, content_model):
        self._slots[slot_key] = {}
        # To avoid TypeError: argument of type 'NoneType' is not iterable in
        #  set_slot_content()
        self._slots_changed[slot_key] = False
        self._content_model[slot_key] = content_model

    def get_content(self):
        return self._content

    def get_slot_content(self, slot_key):
        if slot_key not in self._slots:
            return None
        return self._slots[slot_key]

    def get_slot_content_model(self, slot_key):
        if slot_key not in self._slots:
            return None
        return self._content_model[slot_key]

    def set_content(self, content):
        self._content = content
        self.changed = True

    def set_slot_content(self, slot_key, content):
        if slot_key not in self._slots:
            slot_dict = SLOTS.get(slot_key, None)
            if slot_dict is None:
                raise ValueError(
                    f"Error: Slot '{slot_key}' not defined in 'SLOTS'."
                    f"Available slots: {list(SLOTS.keys())}"
                )
            content_model = slot_dict["content_model"]
            self.create_slot(slot_key, content_model)
        if content != self._slots[slot_key]:
            self._slots_changed[slot_key] = True
        self._slots[slot_key] = content

    def get_url(self) -> str:
        return "https://" + self.wtSite._site.host + "/wiki/" + self.title

    def is_file_page(self) -> bool:
        """Checks if this page is a file page (containing an image, pdf, etc.)

        Returns
        -------
            true if this page is a file page, else false
        """
        return self.title.startswith("File:")

    def append_template(self, template_name: str = None, template_params: dict = None):
        self._dict.append({template_name: template_params})
        return self

    def append_text(self, text):
        self._dict.append(text)
        return self

    def get_value(self, jsonpath):
        jsonpath_expr = parse(jsonpath)
        res = []
        d = dict(
            zip(range(len(self._dict)), self._dict)
        )  # convert list to dict with index
        for match in jsonpath_expr.find(d):
            res.append(match.value)
        return res

    def update_dict(self, combined: dict, update: dict) -> None:
        for k, v in update.items():
            if isinstance(v, dict):
                # todo: fix reference for combine_into
                WtPage.combine_into(v, combined.setdefault(k, {}))
            else:
                combined[k] = v

    def set_value(self, jsonpath_match, value, replace=False):
        jsonpath_expr = parse(jsonpath_match)
        d = dict(
            zip(range(len(self._dict)), self._dict)
        )  # convert list to dict with index
        # if create: jsonpath_expr.update_or_create(d, value)
        # else: jsonpath_expr.update(d, value)
        matches = jsonpath_expr.find(d)
        for match in matches:
            print(match.full_path)
            # pprint(value)
            if not replace:
                WtPage.update_dict(match.value, value)
                value = match.value
            # pprint(value)
            match.full_path.update_or_create(d, value)
        self._dict = list(d.values())  # convert dict with index to list
        return self

    def update_content(self):
        self._content = wt.get_wikitext_from_flat_content_structure(self._dict)
        self.changed = self._original_content != self._content
        return self

    def edit(self, comment: str = None, mode="action-multislot"):
        """Creates / updates the content of all page slots in the wiki site

        Parameters
        ----------
        comment:
            (optional) edit comment for the page history, by default None
        mode:
            (optional) single API call ('action-multislot') or multiple (
            'action-singleslot'), by default 'action-multislot' (faster)
        """
        retry = 0
        max_retry = 5
        while retry < max_retry:
            try:
                return self._edit(comment, mode)
            except Exception as e:
                print(e)
                if retry < max_retry:
                    retry += 1
                    print(f"Page edit failed. Retry ({retry}/{max_retry})")
                    sleep(5)

    def _edit(self, comment: str = None, mode="action-multislot"):
        if not comment:
            comment = "[bot] update of page content"
        if self.changed:
            self._page.edit(self._content, comment)  # legacy mode
        if mode == "action-multislot":
            params = {}
            changed = False
            for slot_key in self._slots:
                if self._slots_changed[slot_key]:
                    changed = True
                    self._slots_changed[slot_key] = False
                    content = self._slots[slot_key]
                    if self._content_model[slot_key] == "json":
                        if not isinstance(content, str):
                            content = json.dumps(content, ensure_ascii=False)
                    params["slot_" + slot_key] = content
            if changed:
                self.wtSite._site.api(
                    "editslots",
                    token=self.wtSite._site.get_token("csrf"),
                    title=self.title,
                    summary=comment,
                    **params,
                )
                self.wtSite._clear_cookies()

        else:
            for slot_key in self._slots:
                if self._slots_changed[slot_key]:
                    content = self._slots[slot_key]
                    if self._content_model[slot_key] == "json":
                        content = json.dumps(content, ensure_ascii=False)
                    self.wtSite._site.api(
                        "editslot",
                        token=self.wtSite._site.get_token("csrf"),
                        title=self.title,
                        slot=slot_key,
                        text=content,
                        summary=comment,
                    )
                    self._slots_changed[slot_key] = False

    def delete(self, comment: str = None):
        self._page.delete(comment)

    def move(self, new_title: str, comment: str = None, redirect=True):
        if new_title != self.title:
            print(f"move '{self.title}' to '{new_title}'")
            self._page.move(
                new_title=new_title, reason=comment, no_redirect=not redirect
            )
            self.title = new_title

    def get_last_changed_time(self):
        return datetime.fromisoformat(
            self._current_revision["timestamp"].replace("Z", "+00:00")
        )

    class PageDumpConfig(model.OswBaseModel):
        """Configuration to dump wiki pages to the file system"""

        target_dir: Union[str, Path]
        """Directory to dump all contents.
        Will be created automatically if not existing"""
        namespace_as_folder: Optional[bool] = True
        """Store page contents in subfolders named according to their namespaces"""
        skip_slot_suffix_for_main: Optional[bool] = False
        """If true, do not include 'main' in the generated content file.
        Useful for pages / wikis using only the main slot."""
        dump_empty_slots: Optional[bool] = False
        """If true, dump all configured slots even empty ones.
        Useful to create initial content in these slots."""
        page_name_as_filename: Optional[bool] = False
        """Use the (human readable) name of a page also for the file naming.
        Useful to identify dumped files manually. The mapping to the page title/id
        (in general a UUID) is ensured anyway through package.json meta data."""

        class Config:
            arbitrary_types_allowed = True  # necessary to allow e.g. np.array as type

    def dump(self, config: PageDumpConfig) -> package.PagePackagePage:
        """Dump this page to the file system

        Parameters
        ----------
        config
            see WtPage.PageDumpConfig

        Returns
        -------
            Metadata of the generated dump
        """
        page_name = self.title.split(":")[-1]
        if ":" in self.title:
            namespace = self.title.split(":")[0]
        else:
            namespace = "Main"

        namespace_const = "NS_" + namespace.upper()
        if namespace in package.NAMESPACE_TO_NAMESPACE_CONST_MAPPING:
            namespace_const = package.NAMESPACE_TO_NAMESPACE_CONST_MAPPING[namespace]

        package_page = package.PagePackagePage(
            name=page_name, namespace=namespace_const, slots={}
        )
        name_in_json_data = False
        if "jsondata" in self._slots and "name" in self._slots["jsondata"]:
            package_page.label = self._slots["jsondata"]["name"]
            name_in_json_data = True

        if config.page_name_as_filename and name_in_json_data:
            # Use name from jsondata as filename for the dump
            dump_name = self._slots["jsondata"]["name"]
        else:
            # Use page name as filename for the dump
            dump_name = page_name

        tar_dir = config.target_dir
        path_prefix = ""
        if config.namespace_as_folder:
            tar_dir = os.path.join(tar_dir, namespace)
            path_prefix = namespace + "/"
        # if not os.path.exists(dir):
        #    os.makedirs(dir)

        def save_to_file(file_path__, content__):
            if not os.path.exists(os.path.dirname(file_path__)):
                os.makedirs(os.path.dirname(file_path__))
            with open(os.path.join(file_path__), "w", encoding="utf-8") as f__:
                f__.write(content__)

        def dump_slot_content(slot_key_, content_type_, content_):
            if isinstance(content_, dict):
                content_ = json.dumps(content_, indent=4, ensure_ascii=False)
            if content_type_ == "Scribunto":
                content_type_ = "lua"
            if slot_key_ == "main" and config.skip_slot_suffix_for_main:
                file_name_ = f"{dump_name}.{content_type_}"
            else:
                file_name_ = f"{dump_name}.slot_{slot_key_}.{content_type_}"
            # handle subpages:
            file_path_ = os.path.join(tar_dir, *file_name_.split("/"))
            save_to_file(file_path_, content_)
            if slot_key_ == "main":
                package_page.urlPath = path_prefix + file_name_
            else:
                package_page.slots[slot_key_] = package.PagePackagePageSlot(
                    urlPath=path_prefix + file_name_
                )

        for slot_key in self._slots:
            content = self.get_slot_content(slot_key)
            content_type = self.get_slot_content_model(slot_key)
            dump_slot_content(slot_key, content_type, content)

        # If the slots are empty, we still want files to fill after dumping them
        if config.dump_empty_slots:
            for slot_key in [slot for slot in SLOTS.keys() if slot not in self._slots]:
                content = SLOTS[slot_key]["content_template"]
                content_type = SLOTS[slot_key]["content_model"]
                dump_slot_content(slot_key, content_type, content)

        if self.is_file_page():
            file = self.wtSite._site.images[self.title.split(":")[-1]]
            file_name = f"{page_name}"
            file_path = os.path.join(tar_dir, *file_name.split("/"))  # handle subpages
            with open(file_path, "wb") as fd:
                file.download(fd)
            package_page.fileURLPath = path_prefix + file_name

        return package_page

    def get_file_info_and_usage(
        self, debug: bool = False
    ) -> Dict[str, Union[str, List[str]]]:
        return wt.get_file_info_and_usage(
            site=self.wtSite._site,
            title=wt.SearchParam(query=self.title, debug=debug),
        )[0]

    def purge(self):
        self._page.purge()


# Updating forwards refs in pydantic models
WtSite.UploadPageParam.update_forward_refs()
WtSite.GetPageResult.update_forward_refs()
WtSite.CreatePagePackageParam.update_forward_refs()
WtSite.UploadPagePackageParam.update_forward_refs()
WtSite.ReadPagePackageResult.update_forward_refs()
